/// @file
////////////////////////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2016/17      Christian Lessig, Otto-von-Guericke Universitaet Magdeburg
///
////////////////////////////////////////////////////////////////////////////////////////////////////
///
///  module     : tutorial 4
///
///  author     : lessig@isg.cs.ovgu.de
///
///  project    : GPU Programming
///
///  description: CUDA basics
///
////////////////////////////////////////////////////////////////////////////////////////////////////

#include <iostream>
#include <algorithm>

#include "cuda_util.h"

////////////////////////////////////////////////////////////////////////////////////////////////////
// multiply vector by constant
////////////////////////////////////////////////////////////////////////////////////////////////////
__global__
void
vecAddC( float* data, const unsigned int n, const float val) {

  unsigned int tid = blockIdx.x * blockDim.x + threadIdx.x;

  if( tid < n) {
    data[tid] += val;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// program entry point
////////////////////////////////////////////////////////////////////////////////////////////////////
int
main( int /*argc*/, char** /*argv*/ ) {

  const unsigned int n = 16;
  float* data = (float*) malloc( sizeof(float) * n);
  for( unsigned int i = 0; i < n; ++i) {
    data[i] = (float) i;
  }

  // check execution environment
  int deviceCount = 0;
  checkErrorsCuda( cudaGetDeviceCount(&deviceCount));
  if( 0 == deviceCount) {
    std::cerr << "initDevice() : No CUDA device found." << std::endl;
    return EXIT_FAILURE;
  }

  // query the device properties
  cudaDeviceProp devProp;
  cudaGetDeviceProperties(&devProp, 0);
  printDeviceProps( devProp);

  // set the device
  int device_handle = 0;
  cudaSetDevice( device_handle);

  // initialize memory
  float* data_device = nullptr;

  // allocate device memory
  checkErrorsCuda( cudaMalloc((void **) &data_device, sizeof(float) * n));

  // copy device memory
  checkErrorsCuda( cudaMemcpy( (void*) data_device, data, sizeof(float) * n, cudaMemcpyHostToDevice ));

  // determine thread layout
  const unsigned int MAX_THREADS_PER_BLOCK = 1024;
  int num_threads_per_block = std::min( n, MAX_THREADS_PER_BLOCK);
  int num_blocks = n / MAX_THREADS_PER_BLOCK;
  if( 0 != n % MAX_THREADS_PER_BLOCK) {
    num_blocks++;
  }
  std::cout << "num_blocks = " << num_blocks << " :: "
            << "num_threads_per_block = " << num_threads_per_block << std::endl;

  // run kernel
  vecAddC<<< num_blocks , num_threads_per_block >>>( data_device, n, 5);
  checkLastCudaError("Kernel execution failed");
  cudaDeviceSynchronize();

  // copy result back to host
  checkErrorsCuda( cudaMemcpy( data, data_device, sizeof(float) * n, cudaMemcpyDeviceToHost ));

  // output
  for( unsigned int i = 0; i < n; ++i) {
    std::cout << "data[" << i << "] = " << data[i] << '\n';
  }

  // clean up device memory
  checkErrorsCuda( cudaFree( data_device));

  // clean up host memory
  free( data);

  return EXIT_SUCCESS;
}
